@@ -2,28 +2,24 @@ from __future__ import annotations
 
 import argparse
 import builtins
-import json
+import enum
 import logging
 import os
 import shlex
-import subprocess as sp
 import sys
-import textwrap
 import uuid
 from functools import partial
 from pathlib import Path
-from typing import Any, cast
 
-import dill  # type: ignore[import]
+import dill
 from kraken.core import Context, Task, TaskGraph
 from slap.core.cli import Command
 from termcolor import colored
 
-from kraken.cli import __version__
 from kraken.cli.buildenv.environment import BuildEnvironment
 from kraken.cli.buildenv.lockfile import Lockfile
 from kraken.cli.buildenv.project import DefaultProjectImpl, ProjectInterface
-from kraken.cli.buildenv.requirements import LocalRequirement
+from kraken.cli.buildenv.requirements import RequirementSpec
 
 DEFAULT_BUILD_DIR = Path("build")
 DEFAULT_PROJECT_DIR = Path(".")
@@ -31,42 +27,22 @@ logger = logging.getLogger(__name__)
 print = partial(builtins.print, flush=True)
 
 
-def get_implied_requirements(develop: bool) -> list[str]:
-    """Returns a list of requirements that are implied for build environments managed by Kraken CLI.
+class InstallSource(enum.Enum):
+    REQUIREMENTS = enum.auto()
+    LOCKFILE = enum.auto()
 
-    :param develop: If set to `True`, it is assumed that the current Kraken CLI is installed in develop mode
-        using `slap link` or `slap install --link` and will be installed from the local project directory on
-        the file system instead of from PyPI. Otherwise, Kraken CLI will be picked up from PyPI.
-    """
+    def readable(self) -> str:
+        return "requirements" if self == InstallSource.REQUIREMENTS else "lock file"
 
-    if develop:
-        import kraken.cli
 
-        init_path = Path(kraken.cli.__file__).resolve()
-        kraken_path = init_path.parent.parent.parent
-        project_root = kraken_path.parent
-        pyproject = project_root / "pyproject.toml"
-        if not pyproject.is_file():
-            raise RuntimeError(
-                "kraken-cli does not seem to be installed in development mode (expected kraken-cli's "
-                'pyproject.toml at "%s")' % pyproject
-            )
+class InstallResult(enum.Enum):
+    UP_TO_DATE = enum.auto()
+    INSTALLED = enum.auto()
+    UPDATED = enum.auto()
+    UPGRADED = enum.auto()
 
-        # TODO (@NiklasRosenstein): It would be nice if we could tell Pip to install kraken-cli in
-        #       development mode, but `pip install -e DIR` does not currently work for projects using
-        #       Poetry.
-        return [f"kraken-cli@{project_root}"]
-
-    # Determine the next Kraken CLI release that may ship with breaking changes.
-    version: tuple[int, int, int] = cast(Any, tuple(map(int, __version__.split("."))))
-    min_version = f"0.{version[1]}.0"
-    if version[0] == 0:
-        # While we're in 0 major land, let's assume potential breaks with the next minor version.
-        max_version = f"0.{version[1]+1}.0"
-    else:
-        max_version = f"{version[0]}.0.0"
-
-    return [f"kraken-cli>={min_version},<{max_version}"]
+    def readable(self) -> str:
+        return " ".join(self.name.split("_")).lower()
 
 
 class BuildAwareCommand(Command):
@@ -117,117 +93,81 @@ class BuildAwareCommand(Command):
         """Returns the implementation that deals with project specific data such as build requirements and
         lock files on disk."""
 
-        develop = os.getenv("KRAKEN_DEVELOP") == "1"
-        implied_requirements = get_implied_requirements(develop)
-        return DefaultProjectImpl(args.project_dir, implied_requirements)
+        return DefaultProjectImpl(args.project_dir, kraken_cli_develop=os.environ.get("KRAKEN_DEVELOP") == "1")
 
-    def install(self, build_env: BuildEnvironment, project: ProjectInterface, upgrade: bool = False) -> None:
+    def install(
+        self,
+        build_env: BuildEnvironment,
+        project: ProjectInterface,
+        force: bool = False,
+    ) -> tuple[InstallSource, InstallResult]:
         """Make sure that the build environment exists and the requirements are installed.
 
         :param build_env: The build environment to ensure is up to date.
         :param project: Implementation that provides access to the requirement spec and lockfile.
-        :param upgrade: If set to `True`, ignore the lockfile and reinstall requirement spec.
+        :param force: If set to `True`, ignore the lock file and existing build environment.
         """
 
-        re_installed = self._install_or_reuse_env(build_env, project, upgrade)
-
-        if not re_installed:
-            requirements = project.get_requirement_spec()
-            local_requirements = [r for r in requirements.requirements if isinstance(r, LocalRequirement)]
-            if os.getenv("KRAKEN_ALWAYS_UPDATE_LOCAL_REQUIREMENTS") == "1" and local_requirements:
-                print(colored("Updating local requirements due to KRAKEN_ALWAYS_UPDATE_LOCAL_REQUIREMENTS=1", "blue"))
-                build_env.update_local_requirements(local_requirements)
-
-    def _install_or_reuse_env(
-        self,
-        build_env: BuildEnvironment,
-        project: ProjectInterface,
-        upgrade: bool = False,
-    ) -> bool:
-        """Installs the environment. Returns True if the entire environment was (re)installed."""
-
-        if not build_env.exists():
-            print(
-                colored(
-                    "Creating Kraken build environment (%s) ..." % (colored(str(build_env.path), attrs=["bold"]),),
-                    "blue",
-                )
-            )
-            build_env.create(None)
-        else:
-            print(
-                colored(
-                    "Reusing Kraken build environment (%s)" % (colored(str(build_env.path), attrs=["bold"]),),
-                    "blue",
-                )
-            )
-
-        # NOTE (@NiklasRosenstein): This requirement spec will already contain the implied Kraken CLI requirement.
         requirements = project.get_requirement_spec()
+        lockfile_path = project.get_lock_file()
+        lockfile = Lockfile.from_path(lockfile_path)
 
-        if not upgrade:
-            lockfile_path = project.get_lock_file()
-            lockfile = Lockfile.from_path(lockfile_path)
-            if lockfile is not None:
-                if lockfile.requirements.to_hash() != build_env.hash:
-                    if build_env.hash is None:
-                        print(colored("Installing from lockfile ...", "blue"))
-                    else:
-                        print(
-                            colored(
-                                "Warning: Your build environment is outdated compared to your lockfile (%s).\n"
-                                "         Reinstalling from lockfile now ..."
-                                % (colored(str(lockfile_path), attrs=["bold"]),),
-                                "yellow",
-                            )
-                        )
-                    build_env.install_lockfile(lockfile)
-                    build_env.hash = lockfile.requirements.to_hash()
+        if force and lockfile:
+            logger.info("Ignoring lockfile (%s)", lockfile_path)
+            lockfile = None
 
-                if lockfile.requirements.to_hash() != requirements.to_hash():
-                    print(
-                        colored(
-                            "Warning: Your lockfile (%s) appears to be outdated. Consider re-writing it using the\n"
-                            "         %s command."
-                            % (
-                                colored(str(lockfile_path), attrs=["bold"]),
-                                colored("kraken env lock", "cyan", attrs=["underline"]),
-                            ),
-                            "yellow",
-                        )
-                    )
-                    prefix = "         "
-                    print()
-                    print(prefix + colored("Requirements in lockfile:", "yellow"))
-                    print(textwrap.indent(json.dumps(lockfile.requirements.to_json(), indent=2), prefix + "  "))
-                    print(prefix + colored("Requirements in project:", "yellow"))
-                    print(textwrap.indent(json.dumps(requirements.to_json(), indent=2), prefix + "  "))
-
-                return False
-
-        if build_env.hash is not None and requirements.to_hash() != build_env.hash:
-            print(
-                colored(
-                    "Warning: Your build environment is outdated compared to your requirements.\n"
-                    "         Reinstalling from requirements now ...",
-                    "yellow",
-                )
+        build_env_existed = build_env.exists()
+        if os.getenv("KRAKEN_REINSTALL_BUILD_ENV") == "1" and build_env.exists():
+            logger.info(
+                "Ignoring existing build environment (%s) because of KRAKEN_REINSTALL_BUILD_ENV=1",
+                build_env.path,
             )
-            install = True
-        elif build_env.hash is None:
-            print(colored("Installing from requirements ...", "blue"))
-            install = True
-        elif upgrade:
-            print(colored("Upgrading environment from requirements ...", "blue"))
-            install = True
+
+        install_required = not build_env.exists()
+        if force and build_env.exists():
+            logger.info("Ignoring existing build environment (%s)", build_env.path)
+            install_required = True
+
+        source: RequirementSpec | Lockfile = requirements
+        source_type = InstallSource.REQUIREMENTS
+
+        if lockfile and lockfile.requirements != requirements:
+            logger.warning("Lock file (%s) is outdated. Consider updating it with `kraken env upgrade`.", lockfile_path)
+
+        if lockfile:
+            logger.info("Using requirements from lock file (%s).", lockfile_path)
+            source = lockfile
+            source_type = InstallSource.LOCKFILE
+            requirements = lockfile.to_pinned_requirement_spec()
         else:
-            install = False
+            # TODO (@NiklasRosenstein): Get path to source of requirements from ProjectImpl
+            logger.info("Using requirements from project (%s).", project.get_requirements_path())
 
-        if install:
-            build_env.install_requirements(requirements, upgrade)
-            build_env.hash = requirements.to_hash()
+        if not install_required and build_env.changed(requirements):
+            logger.warning("Build environment is outdated.")
+            logger.info("  Build environment hashes: %s", ", ".join(build_env.hashes))
+            logger.info("  Requirements hash: %s", requirements.to_hash(build_env.hash_algorithm))
+            install_required = True
 
-        return install
+        if install_required:
+            logger.info("Creating build environment (%s)", build_env.path)
+            build_env.install(requirements)
+            build_env.set_hashes(source)
+            logger.info(
+                "Build environment (%s) installed (hash: %s).",
+                build_env.path,
+                requirements.to_hash(build_env.hash_algorithm),
+            )
+            return source_type, (
+                InstallResult.UPGRADED
+                if build_env_existed and force
+                else InstallResult.UPDATED
+                if build_env_existed
+                else InstallResult.INSTALLED
+            )
+        else:
+            logger.info("Build environment (%s) is up to date.", build_env.path)
+            return source_type, InstallResult.UP_TO_DATE
 
     def dispatch_to_build_environment(self, args: Args) -> int:
         """Dispatch to the build environment."""
@@ -239,21 +179,20 @@ class BuildAwareCommand(Command):
         project = self.get_project_interface(args)
         self.install(build_env, project)
 
-        print(
-            colored(
-                "Dispatching command `%s` to build environment (%s)"
-                % (
-                    "kraken " + " ".join(map(shlex.quote, sys.argv[1:])),
-                    colored(str(build_env.path), attrs=["bold"]),
-                ),
-                "blue",
-            )
+        logger.info(
+            "Dispatching command `%s` to build environment (%s)",
+            "kraken " + " ".join(map(shlex.quote, sys.argv[1:])),
+            build_env.path,
         )
 
-        kraken_cli = build_env.get_program("kraken")
-        env = os.environ.copy()
-        env["KRAKEN_MANAGED"] = "1"
-        return sp.call([str(kraken_cli)] + sys.argv[1:], env=env)
+        with build_env.activate():
+            os.environ["KRAKEN_MANAGED"] = "1"
+            from kraken.cli.__main__ import _main
+
+            try:
+                _main()
+            except SystemExit as exc:
+                return exc.code
 
     def execute(self, args: Args) -> int | None:
         if args.verbose >= 2:
@@ -347,7 +286,7 @@ def load_state(state_dir: Path) -> tuple[Context, TaskGraph] | tuple[None, None]
     state_files = list(state_dir.iterdir()) if state_dir.is_dir() else []
     if not state_files:
         return None, None
-    logger.info("Note: Resuming from %d build state(s)", len(state_files))
+    logger.info("Resuming from %d build state(s)", len(state_files))
     context: Context | None = None
     graph: TaskGraph | None = None
     for state_file in sorted(state_files):
@@ -369,4 +308,4 @@ def save_state(state_dir: Path, graph: TaskGraph) -> None:
     for file in state_dir.iterdir():
         if file != state_file:
             file.unlink()
-    logger.info('Note: Saving build state to "%s"', state_file)
+    logger.info('Saving build state to "%s"', state_file)
